/*
 * @(#)PlayerFile.java
 */
package games.Battle.server.EuropaCore;

import games.Battle.shared.comm.PlayerInfo;
import games.Battle.shared.sys.Logger;

import java.util.*;
import java.io.*;

/**
 * PlayerFile is a thread that manages reading the player file, authorizing
 * players, modifying players, and periodically saving the player file.
 *
 * @version 1.00
 * @author Alex Nicolaou
 * @author Jay Steele
 */

public class PlayerFile implements Runnable {
	/**
	 * the thread that periodically checkpoints the player file
	 */
	Thread checkPoint;
	/**
	 * the log file so that we can note when we checkpoint
	 */
	Logger logfile;

	/**
	 * the actual player file
	 */
	File playerFile;

	/**
	 * a vector of every id that is in the player file.
	 */
	Vector knownIds;
	/** 
	 * checkpointInterval is the delay between writes of the player file.
	 */
	public int checkpointInterval = 600000;

	/**
	 * construct and start the player file thread.
	 * @param l the log file 
	 * @param filename the path to the player file
	 */
	public PlayerFile(Logger l, String filename) throws java.io.IOException {
		playerFile = new File(filename);
		logfile = l;

		readPlayerFile();

		start();
	}

	/**
	 * reads the player file into the knownIds vector.
	 */
	void readPlayerFile() {
		knownIds = new Vector();
		try {
			FileInputStream infile = new FileInputStream(playerFile);
			DataInputStream in = new DataInputStream(infile);

			while (in.available() > 0) {
				String player = in.readLine();
				PlayerInfo id = new PlayerInfo(player);
				knownIds.addElement(id);
			}
		}
		catch (Exception e) {
			logfile.log("reading playerfile: " + e);
			System.exit(1);
		}
		logfile.log("finished reading player file");
	}

	/**
	 * saves the current set of players ot the player file
	 */
	synchronized void checkpoint() {
		try {
			Enumeration e = knownIds.elements();
			FileOutputStream outfile = new FileOutputStream(playerFile);
			DataOutputStream out = new DataOutputStream(outfile);
			while (e.hasMoreElements()) {
				String line = e.nextElement().toString() + "\r\n";
				out.writeBytes(line);
			}
			logfile.log("checkpointed player file");
		}
		catch (Exception e) {
			logfile.log("while checkpointing: " + e);
		}
	}

	/**
	 * Loop forever, sleeping and checkpointing periodically
	 */
    public void run() {
		while (true) {
			try { 
				Thread.sleep(checkpointInterval);
				checkpoint();
			}
			catch (Exception e) {
			}
		}
	}

	/**
	 * start the checkpointing thread
	 */
	public void start() {
		if (checkPoint != null && checkPoint.isAlive())
			return;

		checkPoint = new Thread(this);
		checkPoint.start();
	}

	/**
	 * stop the checkpoint thread
	 */
	public void stop() {
		checkpoint();

		if (checkPoint != null && checkPoint.isAlive())
			checkPoint.stop();
	}

	/**
	 * validate a particular login attempt
	 */
	public boolean authorizePlayer(PlayerInfo id) {
		Enumeration e = knownIds.elements();
		PlayerInfo foundID = null;
		while (e.hasMoreElements()) {
			PlayerInfo test = (PlayerInfo)e.nextElement();
			if (test.sameId(id)) {
				if (test.samePass(id)) {
					return false;
				}
				return true;
			}
		}
		return true;
	}

	/**
	 * get the player info matching a given id
	 */
	public PlayerInfo getPlayerInfo(PlayerInfo id) {
		Enumeration e = knownIds.elements();
		PlayerInfo foundID = null;
		while (e.hasMoreElements()) {
			PlayerInfo test = (PlayerInfo)e.nextElement();
			if (test.sameId(id)) {
				return test;
			}
		}
		return null;
	}

	/**
	 * return true if the name is valid. caveat: this function should be
	 * pickier about what is valid
	 * @param n the name
	 */
	boolean validName(String n) { return n.length() > 0;  }

	/**
	 * create a new player based on id.
	 * @param id the id to create
	 */
	public boolean newPlayer(PlayerInfo id) {
		Enumeration e = knownIds.elements();
		PlayerInfo foundID = null;
		while (e.hasMoreElements()) {
			PlayerInfo test = (PlayerInfo)e.nextElement();
			if (test.sameId(id)) {
				// reject new player if there is any other player with
				// the same id.
				return true;
			}
		}

		if (validName(id.getHandle()) && validName(id.getName())) {
			knownIds.addElement(id);
			return false;
		}

		return true;
	}
}
